from IzVerifier.izspecs.containers.izcontainer import IzContainer
from IzVerifier.izspecs.containers.constants import *


class IzVariables(IzContainer):
    """
    Container for parsing and storing IzPack variables from variables.xml.
    """

    properties = {
        NAME: "variables",
        DEFINITION_SPEC_FILES: ['variables.xml'],
        REFERENCE_SPEC_FILES: ["install",
                               "userInputSpec",
                               "ProcessPanel.Spec",
                               "core-packs",
                               "conditions",
                               "variables",
                               "dynamic_variables"],
        ATTRIBUTES: ['variable'],
        SPEC_ELEMENT: '<variable name=\"{0}\" value=\"{1}\"/>',
        PARENT_OPENING_TAG: '<xfragment>',
        PARENT_CLOSING_TAG: '</xfragment>',
        WHITE_LIST: ['izpack.linuxinstall', 'JAVA_HOME', 'INSTALL_PATH', 'IzPanel.LayoutType',
                     'APP_NAME', 'APP_URL', 'APP_VER', 'INSTALL_GROUP', 'UNINSTALLER_CONDITION',
                     'USER_NAME', 'USER_HOME', 'ISO3_LANG', 'IP_ADDRESS', 'HOST_NAME',
                     'FILE_SEPARATOR', 'DesktopShortcutCheckboxEnabled', 'InstallerFrame.logfilePath'],
        PATTERNS: [('getVariable\({0}', 'getVariable\(({0})\)'),
                   ('setVariable\({0}', 'setVariable\(({0}),({0})\)')],
        READ_PATTERNS: [('getVariable({0}', 'getVariable\(({0})\)')],
        WRITE_PATTERNS: [('setVariable({0}', 'setVariable\(({0}),({0})\)')],
        WHITE_LIST_PATTERNS: []
    }

    def parse(self, soup):
        """
        Extracts all variable definitions from variables.xml doc.
        """
        variables = soup.find_all('variable')
        for var in variables:
            self.container[str(var['name'])] = var

    def get_keys(self):
        """
        Returns a set of all the keys for defined variables.
        """
        return set(self.container.keys()) | set(self.properties[WHITE_LIST])

    def count(self):
        """
        Return number of vars found in definition file.
        """
        return len(self.container.keys())

    def get_spec_elements(self):
        """
        Returns a set of xml elements defining each variable.
        """
        return set(self.container.values())

    def has_reference(self, element):
        """
        Return true if the given element contains an izpack var reference.
        """
        for atty in self.properties['attributes']:
            if element.has_attr(atty):
                return True

        return False

    def to_string(self):
        return str(self.container)

    @staticmethod
    def print_keys():
        """
        Prints all of the variable keys found in definition spec.
        """
        pass

    @staticmethod
    def element_sort_key(element):
        """
        Returns the key to use when sorting elements of this container.
        """
        return element['name'].lower()

    @staticmethod
    def get_identifier(element):
        """
        Returns the identifying value for this element.
        """
        return element['name']

    @staticmethod
    def get_value(element):
        """
        Returns the main 'value' for this element.
        In the case of variables, its value is its value.
        """
        return element['value']

    @staticmethod
    def ref_transformer(ref):
        """
        Wraps reference into list.
        """
        return [ref]

    @staticmethod
    def has_definition(self, element):
        """
        Determines if element is a string definition.
        """
        return 'str' == element.name and element.has_attr('key') and element.has_attr('id')




__author__ = 'fcanas'

import unittest
from IzVerifier.izspecs.containers.izconditions import IzConditions
from IzVerifier.izspecs.containers.izstrings import IzStrings
from IzVerifier.izspecs.containers.izvariables import IzVariables
from IzVerifier.izspecs.verifiers.seeker import Seeker
from IzVerifier.izverifier import IzVerifier
from IzVerifier.izspecs.izproperties import *


path1 = '' # path to specs
path2 = '' # path to resources
path3 = '' # path to pom
source_paths = [''] # paths to code bases
class TestProduct(unittest.TestCase):
    """
    This is a template class used to test the verifier on real izpack projects.
    """

    args = {
            'specs_path': path1,
            'sources': source_paths,
            'resources_path': path2,
            'pom': path3,
            'specs': ['conditions', 'strings', 'variables']
        }

    def setUp(self):
        #self.loadInstaller(self.args)
        pass



    def loadInstaller(self, args):
        """
        Runs the verifications on some installer product.
        """

        self.verifier = IzVerifier(args)
        self.seeker = Seeker(self.verifier.paths)
        self.conditions = IzConditions(self.verifier.paths.get_path('conditions'))
        self.variables = IzVariables(self.verifier.paths.get_path('variables'))
        langpack = self.verifier.paths.get_langpack_path()
        self.strings = IzStrings(langpack)

    def verifyInstaller(self):
        """
        Run verification tests.
        """
        #self.verifier.verify_all(verbosity=1)
        self.verifier.containers.get('classes').print_keys()
        self.verifier.verify('classes', verbosity=2)

    def test_verifyInstaller(self):
        #self.verifyInstaller()
        pass








from IzVerifier.izspecs.containers.izcontainer import IzContainer
from IzVerifier.izspecs.containers.constants import *
from os import walk

__author__ = 'fcanas'


class IzClasses(IzContainer):
    """
    Container for parsing and storing custom classes used in izpack installers.
    """

    properties = {
        NAME: "classes",
        DEFINITION_SPEC_FILES: [],
        REFERENCE_SPEC_FILES: [
            "install",
            "userInputSpec",
            "ProcessPanel.Spec",
            "core-packs"],
        ATTRIBUTES: ['class', 'name', 'classname', 'installer'],
        SPEC_ELEMENT: '',
        PARENT_OPENING_TAG: '',
        PARENT_CLOSING_TAG: '',
        WHITE_LIST: [],
        PATTERNS: [],
        READ_PATTERNS: [],
        WRITE_PATTERNS: [],
        WHITE_LIST_PATTERNS: ['^com.izforge.izpack.*$']
    }

    def __init__(self, path=None):
        """
        Initializes the container from the path to the root of custom source code.
        Note: does not make use of parent class __init__.
        """
        self.container = {}
        self.referenced = {}
        if path:
            self.parse(path)

    def parse(self, root):
        """
        Izclasses are not pre-defined anywhere. All we can do is make a collection of
        all custom classes used in source code and index that.
        """
        for paths, dirs, files in walk(root):
            for f in files:
                if '.java' in f:
                    path = paths + '/' + f
                    name = self.path_to_id(root, path)
                    self.container[name] = path

    def get_keys(self):
        """
        Returns a set of all the keys for defined variables.
        """
        return set(self.container.keys()) | set(self.properties[WHITE_LIST])

    def count(self):
        """
        Return number of vars found in definition file.
        """
        return len(self.container.keys())

    def print_keys(self):
        """
        Prints all of the variable keys found in definition spec.
        """
        for key in self.container.keys():
            print key

    def get_spec_elements(self):
        """
        Returns a set of xml elements defining each variable.
        """
        return set(self.container.values())

    def to_string(self):
        return str(self.container)

    def has_reference(self, element):
        """
        Return true if the given element contains an izpack var reference.
        """

        def is_izpack_class(classname):
            """
            Determines if this references an izpack built-in class.
            """
            if type(classname) is list:
                classname = classname[0]
            return not '.' in classname

        if element.has_attr('name') and element.name == 'executeclass':
            return not is_izpack_class(element['name'])
        if element.has_attr('class'):
            return not is_izpack_class(element['class'])
        if element.has_attr('classname'):
            return not is_izpack_class(element['classname'])
        if element.has_attr('installer') and element.name == 'listener':
            return not is_izpack_class(element['installer'])

        return False

    def has_definition(self, element):
        """
        Custom izpack classes are not defined by xml descriptors, so this method is unused.
        """
        pass

    @staticmethod
    def path_to_id(root, path):
        """
        Transforms a classpath to a class id.
        """
        return path.replace(root, '').replace('/', '.').replace('.java', '')

    @staticmethod
    def element_sort_key(element):
        """
        Returns the key to use when sorting elements of this container.
        """
        return element['class'].lower()

    @staticmethod
    def get_identifier(element):
        """
        Returns the identifying value for this element.
        """
        return element['class']

    @staticmethod
    def get_value(element):
        """
        Returns the main 'value' for this element.
        """
        return element

    @staticmethod
    def ref_transformer(ref):
        """
        Unwraps the ref if necessary.
        """
        return [ref]



__author__ = 'fcanas'

import unittest

from IzVerifier.izverifier import IzVerifier


path1 = 'data/sample_installer_iz5/izpack/'
path2 = 'data/sample_installer_iz5/resources/'
source_path2 = 'data/sample_code_base'
pom = 'data/sample_installer_iz5/pom.xml'



class TestDependencies(unittest.TestCase):
    """
    Basic testing of verifier class.
    """

    def setUp(self):
        args = {
            'specs_path': path1,
            'sources': [source_path2],
            'resources_path': path2,
            'pom': pom,
        }
        self.izv = IzVerifier(args)

    def test_verifyAllDependencies(self):
        """
        Run the full dependency verification test.
        """
        undefined_deps = {'and.1', 'and.2', 'and.3',
                          'or.cycle.1', 'or.cycle.2', 'or.cycle.3',
                          'some.condition.1', 'some.condition.2',
                          'variable1',
                          'myinstallerclass.condition', 'short.1'}

        results = self.izv.dependency_verification(verbosity=2, fail_on_undefined_vars=True)

        extras_found = set(results.keys()) - undefined_deps
        not_found = undefined_deps - set(results.keys())

        self.assertFalse(extras_found)
        self.assertFalse(not_found)

        for cond in extras_found:
            print "%s should not have been found" % cond

        for cond in not_found:
            print "%s was not found" % cond











__author__ = 'fcanas'
PARENT_CLOSING_TAG = 'parent_closing_tag'
PARENT_OPENING_TAG = 'parent_opening_tag'
PATTERNS = 'patterns'
SPEC_ELEMENT = 'spec_element'
ATTRIBUTES = 'attributes'
CHECK_KEY_PATTERN = 'check_key_pattern'
SOURCE_SEARCH_PATTERN = 'source_search_pattern'
REFERENCE_SPEC_FILES = 'reference_spec_files'
DEFINITION_SPEC_FILES = 'definition_spec_files'
WRITE_PATTERNS = 'write_patterns'
READ_PATTERNS = 'read_patterns'
WHITE_LIST = 'white_list'
NAME = 'name'
WHITE_LIST_PATTERNS = 'white_list_patterns'
KEY_PATTERN = 'get_key_pattern'
REF_ID_ATTRIBUTES_LIST = 'ref_id_attributes_list'
IGNORE_KEY_PATTERN = 'ignore_key_pattern'




__author__ = 'fcanas'

import termhelper


class Reporter:
    """
    Responsible for displaying results and recording them to log files.
    """

    def __init__(self, warg=0):
        self.width = warg
        self.set_terminal_width(warg)  # default width of terminal

    templates = {
        'test': '[ {0:.<{2}}{1:.>{3}} ]',
        'set_tuples': '{0:.<{2}}{1:.>{3}}',
        'set_items': '{0:.<{1}}',
        'path_display': '[ {0:.<{1}}FAIL ]'
    }

    def report_test(self, test, items):
        """
        Report header for specification results.
        """
        template = self.templates['test']
        test_width = self.width - 4  # Test tuple is wrapped in [  ], which is 4 characters
        print template.format(test, len(items), 0, test_width - len(test), len(test))
        self.report_set(items)

    def report_set(self, entities):
        """
        Human-readable report for a set of unverified entities.
        """
        for item in sorted(entities):
            if type(item) is tuple:
                template = self.templates['set_tuples']
                tuple_padding = self.get_tuple_padding(item)
                print template.format(item[0], item[1], tuple_padding, len(item[1]))
            else:
                template = self.templates['set_items']
                print template.format(item, self.width)

    def set_terminal_width(self, warg=-1):
        """
        Overwrites default terminal width with the width of the current terminal window or the width arg passed
        in by user (if any and positive).
        """
        if warg <= 0:
            height, width = termhelper.terminal_height_width()
            self.width = max(width, 100)

    def get_tuple_padding(self, item):
        """
        Returns the proper length of padding for a tuple.
        The first item is allocated at least 30 characters of space.
        The second item is allocated up to 50 characters.
        """

        # Default padding, used when both items fit in their own allocated space without shifting the padding
        # and when a line break is unavoidable (default padding keeps the output more readable than no padding
        # in the case of a very short first item with a very long second item)

        tuple_padding = max(30, self.width - 50, len(item[0]) + 1)

        # If the combined length of the items is short enough to fit on one line, avoid a line break
        if len(item[0]) + len(item[1]) < self.width:

            # If the first item fills up its allocated space, pad by 1 character
            if len(item[0]) >= max(30, self.width - 50):
                tuple_padding = len(item[0]) + 1

            # If the second item is longer than its allocated space, shorten the padding to avoid a line break
            elif len(item[1]) > min(self.width - 30, 50):
                tuple_padding = self.width - len(item[1])

        return tuple_padding


    def display_paths(self, paths_dict):
        """
        Human readable output for displaying dependency paths.
        """
        def node_type(node):
            return " (type: {0})".format(str(node[1]))

        template = self.templates['path_display']

        for condition_id in paths_dict:
            for path_index, path in enumerate(list(paths_dict[condition_id])):
                tab = len(condition_id)
                for node_index, node in enumerate(path):
                    if node_index == 0:
                        if path_index == 0:
                            print template.format(condition_id + node_type(node), self.width - len('[ FAIL ]'))
                        else:
                            continue
                    else:
                        add_to_tab = 0
                        if type(node[0]) is tuple:
                            cid = node[0][0]
                            add_to_tab += len(cid)
                        else:
                            cid = node[0]
                            add_to_tab += len(cid)
                        if tab:
                            branch = u'\u02ea\u2192 depends on '
                            add_to_tab += len(branch)
                        else:
                            branch = ''

                        print " " * tab + branch + str(cid) + node_type(node)
                        tab += add_to_tab
            print


import os
from os.path import dirname
from bs4 import BeautifulSoup
import re
from IzVerifier.exceptions.IzVerifierException import MissingFileException

__author__ = 'fcanas'


class IzPaths():
    """
    Class responsible for providing paths to specific IzPack resources and spec files.
    """
    
    def __init__(self, specs, resources, properties=None):
        """
        Initialize the installer's root path.
        """
        if not properties:
            properties = {}
        self.init_collections()
        self.properties = properties
        self.set_paths(specs, resources)
        self.parse_paths()
        self.find_resources()

    def init_collections(self):
        """
        Initialize collections to their default values.
        """
        # Default paths to spec files relative to specs folder.
        self.specs = {
            'BASE': '',
            'variables': 'variables.xml',
            'conditions': 'conditions.xml',
            'dynamicvariables': 'dynamic_variables.xml',
            'dynamic_variables': 'dynamic_variables.xml',
            'resources': 'resources.xml',
            'panels': 'panels.xml',
            'packs': 'packs.xml',
            'install': 'install.xml',
            'ProcessPanel.Spec': 'ProcessPanel.Spec.xml',
            'core-packs': 'packs.xml'
        }

        # Default paths to resource files relative to specs folder.
        self.resources = {
            'BASE': '',
            'userInputSpec': 'userInputSpec.xml',
            'strings': 'CustomLangPack.xml'
        }

        self.langpacks = {}

    def set_paths(self, specs, resources):
        """
        Takes base paths to specs and resources.
        """
        self.install = 'install.xml'
        self.specs_path = path_format(specs)
        self.root = path_format(dirname(dirname(self.specs_path)) + '/')
        self.res_path = path_format(resources)
        self.resources['BASE'] = self.res_path
        self.specs['BASE'] = self.specs_path

    def parse_paths(self):
        """
        Extracts paths to available izpack resources and spec files from the
        installer's install.xml spec.
        """
        self.soup = BeautifulSoup(open(self.get_path('install')))
        for spec in self.specs.keys():
            spec_file = self.find_specs_path(spec)
            if spec_file:
                # If spec file exists
                self.specs[spec] = path_format(spec_file)
            else:
                # If specs are held inside install.xml
                self.specs[spec] = self.install

    def find_specs_path(self, spec):
        """
        Find the path for the spec in the install.xml file.
        """
        element = self.soup.find(spec)
        if element:
            child = element.find('xi:include')
            if child:  # if xi:include exists, specs are external.
                path = self.properties.substitute(child['href'])
            else:
                # Internal specs.
                path = None
        else:
            # No spec defined in file, assume default location.
            path = self.specs[spec]
        return path

    def get_path(self, name):
        """
        Returns a path to the spec or resources file, or None if there isn't any.
        """
        for col in [self.specs, self.resources]:
            if name in col:
                return force_absolute(col['BASE'], col[name])
        raise MissingFileException(name)

    def find_resources(self):
        """
        Parse the install.xml resources and extract paths to available resource files.
        """
        path = self.get_path('resources')

        if not path:
            rsoup = self.soup
        else:
            rsoup = BeautifulSoup(open(path))

        self.parse_resources(rsoup)

    def parse_resources(self, soup):
        """
        Parse the install.xml (or resources.xml) soup to find all available resources.
        """
        for res in soup.find_all('res'):
            if 'customlangpack' in res['id'].lower():
                self.find_langpack_path(res)
            else:
                rid = remove_xml(res['id'])
                self.resources[rid] = path_format(self.properties.substitute(res['src']))

    def find_langpack_path(self, langpack):
        """
        Finds a langpack path from the given xml langpack element
        """
        lid = langpack['id']
        src = path_format(self.properties.substitute(langpack['src']))
        if '.xml_' in lid:
            self.langpacks[lid[-3:]] = src

        if not 'default' in self.langpacks:
            self.langpacks['default'] = src
            self.resources['strings'] = src

    def get_langpacks(self):
        """
        Returns a dict of found langpacks in the form:
        {
            'id', 'src'
            ...
        }
        """
        return self.langpacks

    def get_langpack_path(self, lid='default'):
        """
        Returns the path to the langpack with the given localization id.
        """
        path = self.langpacks[lid]
        return force_absolute(self.res_path, path)


def remove_xml(rid):
    """
    Removes the .xml from a resource or spec id.
    """
    if '.xml' in rid[-4:]:
        return rid[:-4]
    else:
        return rid


def force_absolute(base, path):
    """
    Ensures that the base path is not appended to an absolute path.
    """
    if base in path:
        return path
    else:
        return path_format(base + path)


def path_format(path):
    """
    Ensures the path is formatted correctly:

    Folder ends in a '/', and a path to a file does not.

    Removes any double '//'.
    """
    if os.path.isdir(path):
        if path[-1] != '/':
            path += '/'
    return re.sub(r'/+/', '/', path)






import sys
import fcntl
import termios
import struct

def terminal_height_width():
    """
    Attempt to determine size of terminal window, and set to 25x80 if an exception is thrown.
    """
    try:
        height, width = struct.unpack('hh',  fcntl.ioctl(sys.stdout, termios.TIOCGWINSZ, '1234'))
    except IOError:
        height, width = 25, 80

    return height, width

if __name__ == "__main__":
    term_height, term_width = terminal_height_width()
    print "The terminal has", term_height, "rows and", term_width, "columns"



from IzVerifier.izspecs.izpaths import path_format

__author__ = 'fcanas'

import unittest

pom = 'data/sample_installer_iz5/pom.xml'
path1 = 'data/sample_installer_iz5/izpack/'
path2 = 'data/sample_installer_iz5/resources/'
source_path2 = 'data/sample_code_base'

class TestSeeker(unittest.TestCase):
    """
    Basic testing of izpaths class.
    """

    def test_pathFormat(self):
        """
        Tests the path formatting function.
        """
        p1 = 'data//sample_installer_iz5'
        r1 = path_format(p1)
        self.assertEquals('data/sample_installer_iz5/', r1)

        p1 = 'data//sample_installer_iz5/pom.xml'
        r1 = path_format(p1)
        self.assertEquals('data/sample_installer_iz5/pom.xml', r1)

        p1 = 'data///sample_installer_iz5/pom.xml'
        r1 = path_format(p1)
        self.assertEquals('data/sample_installer_iz5/pom.xml', r1)




from bs4 import BeautifulSoup

from IzVerifier.izspecs.containers.izcontainer import IzContainer
from IzVerifier.izspecs.containers.constants import *


IZPACK_LANG_PATH = 'izpack_lang_path'


class IzStrings(IzContainer):
    """
    Container for izpack strings from langpack files.
    """
    properties = {
        NAME: 'strings',
        DEFINITION_SPEC_FILES: ['eng.xml'],
        REFERENCE_SPEC_FILES: ["install",
                               "userInputSpec",
                               "ProcessPanel.Spec",
                               "core-packs"],
        SOURCE_SEARCH_PATTERN: 'langpack.getString(',
        CHECK_KEY_PATTERN: 'langpack.getString\(\"(.*?)\"\)',
        ATTRIBUTES: ['id', 'tooltip', 'variable'],
        SPEC_ELEMENT: '<str id=\"{0}\" txt=\"{1}\"/>',
        PATTERNS: [
            ('langpack.getString\({0}', 'langpack.getString\(({0})\)'),
            ('setError\({0}', 'setError\(({0})\)'),
            ('setMessage\({0}', 'setMessage\(({0})\)'),
            ('System.*.println\({0}', 'System.*.println\(({0})\)'),
            ('setErrorMessageId\({0}', 'setErrorMessageId\(({0})\)')],
        WHITE_LIST_PATTERNS: ['^.*\(\(String\) conn\);.*$',
                              '^UserInputPanel.*$'],  # for some weirdness in a console helper :)

        PARENT_OPENING_TAG: '<langpack>',
        PARENT_CLOSING_TAG: '</langpack>',
    }

    def parse_izpack_strings(self, path):
        """
        Parse izpack's built-in langpack strings.
        """
        self.soup_izpack = BeautifulSoup(open(path), 'xml')
        self.parse(self.soup_izpack)

    def parse(self, soup):
        """
        Extracts all strings from the soup document.
        """
        strings = soup.find_all(self.has_definition)
        for st in strings:
            self.container[str(st['id'])] = st

    def has_definition(self, element):
        """
        Return true if the given element contains an izpack string definition.
        """
        return str(element.name) == 'str' and element.has_attr('id') and element.has_attr('txt')

    def has_reference(self, element):
        """
        Return true if the given element contains an izpack string reference.
        This method is used to define all the rules that allow the verifier to find
        references to the type of iz entity being searched for.
        """

        # panels never contain strings directly as attributes. Their 'id' is the
        # panel id.
        if str(element.name) == 'panel':
            return False

        # Validators directly under panels have id's that are not strings.
        if str(element.parent.name) == 'panel' and str(element.name) == 'validator':
            return False

        # Elements with ids that start with the word 'port' in them are not strings either.
        if element.has_attr('id'):
            if element['id'].startswith('port') or 'maximum.offset.variable' in element.name:
                return False

        # for special case variables that are shown in summary panel and therefore require
        # their own eng string key:
        if element.name == 'field' and element.has_attr('variable') and not element['type'] == 'rule':
            if element.has_attr('summarize') and 'false' in element['summarize']:
                return False
            if element.has_attr('autoPrompt') and 'true' in element['autoPrompt']:
                return False
            return True

        # rule elements don't use strings
        if element.has_attr('type') and 'rule' in element['type']:
            return False

        if 'executeForPack' in element.name and element.has_attr('id'):
            return False

        # Otherwise, check if this element contains one of the attributes used to
        # hold string ids and return true if it does.
        for atty in IzStrings.properties['attributes']:
            if element.has_attr(atty):
                return True

        return False

    def get_keys(self):
        """
        """
        return self.container.keys()

    def count(self):
        """
        """
        return len(self.container.keys())

    def get_spec_elements(self):
        """
        Returns a set of xml elements defining each condition.
        """
        return set(self.container.values())

    @staticmethod
    def ref_transformer(reference):
        """
        Identity function.
        """
        return reference

    @staticmethod
    def element_sort_key(element):
        """
        Returns the key to use when sorting elements of this container.
        """
        return element['id'].lower()



import re

from IzVerifier.izspecs.containers.izcontainer import IzContainer
from IzVerifier.izspecs.containers.constants import *


class IzConditions(IzContainer):
    """ Container for parsing and storing IzPack conditions from conditions.xml. """

    properties = {
        NAME: "conditions",
        DEFINITION_SPEC_FILES: ['conditions.xml'],
        REFERENCE_SPEC_FILES: ["install",
                               "userInputSpec",
                               "ProcessPanel.Spec",
                               "core-packs"],
        IGNORE_KEY_PATTERN: '^izpack.*',  # used by izpack internal vars
        SOURCE_SEARCH_PATTERN: 'isConditionTrue(',
        CHECK_KEY_PATTERN: "isConditionTrue\(\"(.*?)\"\)",
        KEY_PATTERN: "getCondition\(\"(.*?)\"\)",
        ATTRIBUTES: ['condition', 'conditionid', 'refid'],
        REF_ID_ATTRIBUTES_LIST: ['refid'],
        SPEC_ELEMENT: "<condition type=\"{0}\" id=\"{1}\">",
        PARENT_OPENING_TAG: '<conditions>',
        PARENT_CLOSING_TAG: '</conditions>',
        WHITE_LIST: ['izpack.linuxinstall'],
        WHITE_LIST_PATTERNS: [],
        PATTERNS: [('isConditionTrue\({0}', "isConditionTrue\(({0})\)"),
                   ("getCondition\({0}", "getCondition\(({0})\)")]
    }

    def parse(self, soup):
        """ Extracts all conditions from conditions.xml document. """
        conds = soup.find_all(self.has_condition_definition)
        for cond in conds:
            self.container[str(cond['id'])] = cond

    def get_keys(self):
        """
        Returns a set of all the keys for defined conditions.
        """
        return set(self.container.keys()) | set(self.properties[WHITE_LIST])

    def referenced_variables(self):
        """
        Returns the ids for all variables referenced by variable type conditions.
        """
        variables = set()
        for cond in self.container.itervalues():
            if self.has_def_by_variable_ref(cond):
                if cond.find('name'):
                    var = cond.find('name').get_text()
                    variables.add((var, cond['id']))
                else:
                    variables.add(("no_var_defined", cond['id']))
        return variables

    def print_keys(self):
        for cond in self.container.keys():
            print(cond)

    def count(self):
        return len(self.container.keys())

    def get_spec_elements(self):
        """
        Returns a set of xml elements defining each condition.
        """
        return set(self.container.values())

    def to_string(self):
        return str(self.container)

    def ref_transformer(self, reference):
        """
        Given a compound condition id returns a list of the conditions contained in it.
        Example:. "cond1+cond2+!cond3" => {cond1, cond2, cond3}
        """
        cids = re.split("\+|\|", reference)
        return set((cid.replace("!", "") for cid in cids if
                    not re.match(self.properties['ignore_key_pattern'], cid.replace("!", ""))))

    def has_reference(self, element):
        """
        Return true if the given element contains an izpack cond reference.
        """
        for atty in self.properties['attributes']:
            if element.has_attr(atty):
                return True

        return False

    def has_definition(self, element):
        """
        Determines if the given xml element is a condition definition.
        """
        return 'condition' == element.name and element.has_attr('type') and element.has_attr('id')

    @staticmethod
    def has_def_by_condition_ref(ele):
        """
        Function to determine if element contains a ref id for conditions.
        """
        return ele.has_attr('refid')

    @staticmethod
    def has_condition_definition(element):
        """
        Returns true if this element is a condition definition.
        """
        return element.name == 'condition' and element.has_attr('id')

    @staticmethod
    def has_def_by_variable_ref(ele):
        """
        Does this element define a 'variable' type condition?
        """
        return ele.has_attr('type') and ele['type'] == 'variable'

    @staticmethod
    def extract_variable_from_definition(element):
        """
        Given an element containing a variable type condition, return the variable name.
        """
        return element.select('name').get_text()

    @staticmethod
    def get_identifier(element):
        """
        Returns the identifying value for this element.
        """
        return element['id']

    @staticmethod
    def get_value(element):
        """
        Returns the main 'value' for this element.
        In the case of conditions, its value is the entire definition, or element.
        """
        return element

    @staticmethod
    def element_sort_key(element):
        """
        Returns the key to use when sorting elements of this container.
        """
        return element['id'].lower()





from bs4 import BeautifulSoup
import re

__author__ = 'fcanas'


class IzProperties(dict):
    """
    Responsible for parsing and containing any properties used by IzPack's installation spec files.
    """
    def __init__(self, path):
        """
        Initialize paths to properties and begin parsing.
        """
        # noinspection PyTypeChecker
        dict.__init__(self)
        if 'pom.xml' in path:
            self.parse_pom_properties(path)
        else:
            self.parse_properties(path)

    def parse_properties(self, path):
        """
        Finds properties defined in properties file at specified path adds them to map.
        """
        soup = BeautifulSoup(open(path, 'r'))
        properties = soup.find_all('properties')

        for props in properties:
            for prop in props.find_all('property'):
                try:
                    self[prop['name']] = prop['value']
                except KeyError:
                    continue

    def parse_pom_properties(self, path):
        """
        Special parser for pom.xml file properties.
        """
        soup = BeautifulSoup(open(path, 'r'), 'xml')
        properties = soup.find_all('properties')

        # add the basedir property
        self['basedir'] = path.replace('pom.xml', '')

        for props in properties:
            for prop in props.find_all(recursive=False):
                try:
                    self[str(prop.name)] = str(prop.string)
                except KeyError:
                    continue

    def substitute(self, string):
        """
        Puts the given string through variable substitution: replacing all incidences of
        ${key} for the key's value if it exists. If key doesn't exist, it returns
        the unsubstituted variable. The substitution is performed iteratively until all
        possible variables have been subbed.
        """
        while True:
            old_string = string
            matches = re.findall('\$\{.*\}', string)

            if not matches:
                break

            for match in matches:
                value = self._substitute(match)
                if not value is match:
                    string = str.replace(string, match, value)

            if string is old_string:
                break
        return string

    def _substitute(self, key):
        """
        Substitutes a given key for its value. If the value doesn't exist,
        return the key.

        Key is in the form ${some.key}
        """
        stripped_key = key[2:-1]
        if stripped_key in self:
            return self[stripped_key]
        else:
            return key



from IzVerifier.izspecs.izproperties import IzProperties

__author__ = 'fcanas'

import unittest

from IzVerifier.izspecs.containers.izconditions import IzConditions
from IzVerifier.izspecs.containers.izstrings import IzStrings
from IzVerifier.izspecs.containers.izvariables import IzVariables
from IzVerifier.izspecs.verifiers.seeker import Seeker
from IzVerifier.izverifier import IzVerifier
from IzVerifier.izspecs.containers.constants import *


path1 = 'data/sample_installer_iz5/izpack/'
path2 = 'data/sample_installer_iz5/resources/'
source_path2 = 'data/sample_code_base'
pom = 'data/sample_installer_iz5/pom.xml'


class TestSeeker(unittest.TestCase):
    """
    Basic testing of seeker class.
    """

    def setUp(self):
        args = {
            'specs_path': path1,
            'sources': [source_path2],
            'resources_path': path2,
            'pom': pom,
            'specs': ['conditions', 'strings', 'variables']
        }
        self.verifier = IzVerifier(args)
        self.seeker = Seeker(self.verifier.paths)
        langpack = self.verifier.paths.get_langpack_path()
        self.conditions = IzConditions(self.verifier.paths.get_path('conditions'))
        self.variables = IzVariables(self.verifier.paths.get_path('variables'))
        self.strings = IzStrings(langpack)

    def test_findStringReference(self):
        """
        Find specified string reference across some spec files.
        """
        props = {
            'path': self.verifier.paths.root,
            'id': 'some.user.panel.title',
            'specs': ['data/sample_installer_iz5/izpack/install.xml',
                      'data/sample_installer_iz5/resources/userInputSpec.xml'],
            'filter_fn': self.strings.has_reference,
            'attributes': self.strings.properties[ATTRIBUTES],
            'transformer': lambda x: x,
            'patterns': self.strings.properties[PATTERNS],
            'source_paths': [],
            'white_list_patterns': self.strings.properties[WHITE_LIST_PATTERNS]
        }
        results = self.seeker.find_id_references(props)
        hits = len(results)
        self.assertEquals(hits, 1, msg=str(hits) + '!=1')

    def test_findConditionReference(self):
        """
        Find specified condition reference across some spec files.
        """

        props = {
            'path': self.verifier.paths.root,
            'id': 'some.condition.1',
            'specs': map(self.verifier.paths.get_path, self.conditions.properties[REFERENCE_SPEC_FILES]),
            'filter_fn': self.conditions.has_reference,
            'attributes': self.conditions.properties[ATTRIBUTES],
            'transformer': lambda x: x,
            'patterns': self.conditions.properties[PATTERNS],
            'source_paths': [],
            'white_list_patterns': self.conditions.properties[WHITE_LIST_PATTERNS]
        }
        results = self.seeker.find_id_references(props)
        hits = len(results)
        self.assertEquals(hits, 1, msg=str(hits) + '!=1')

    def test_findVariableReference(self):
        """
        Find specified variable reference across some spec files.
        """

        props = {
            'path': self.verifier.paths.root,
            'id': 'some.user.password',
            'specs': map(self.verifier.paths.get_path, self.variables.properties[REFERENCE_SPEC_FILES]),
            'filter_fn': self.variables.has_reference,
            'attributes': self.variables.properties[ATTRIBUTES],
            'transformer': lambda x: x,
            'patterns': self.variables.properties[PATTERNS],
            'source_paths': [],
            'white_list_patterns': self.strings.properties[WHITE_LIST_PATTERNS]
        }
        results = self.seeker.find_id_references(props)
        hits = len(results)
        self.assertEquals(hits, 1, msg=str(hits) + '!=1')

    def test_findAllStringsInSpecs(self):
        """
        Finds all strings referenced in installer spec files.
        """

        props = {
            'path': self.verifier.paths.root,
            'specs': map(self.verifier.paths.get_path, self.strings.properties[REFERENCE_SPEC_FILES]),
            'filter_fn': self.strings.has_reference,
            'attributes': self.strings.properties[ATTRIBUTES],
        }

        hits = self.seeker.search_specs_for_attributes(props)
        self.assertTrue(len(hits) == 7, msg=str(len(hits)) + '!=7')

    def test_findAllConditionsInSpecs(self):
        """
        Finds all strings referenced in installer spec files.
        """

        props = {
            'path': self.verifier.paths.root,
            'specs': map(self.verifier.paths.get_path, self.conditions.properties[REFERENCE_SPEC_FILES]),
            'filter_fn': self.conditions.has_reference,
            'attributes': self.conditions.properties[ATTRIBUTES],
            'white_list_patterns': []
        }

        hits = self.seeker.search_specs_for_attributes(props)
        self.assertTrue(len(hits) == 4)


    def test_findAllVariablesInSpecs(self):
        """
        Finds all strings referenced in installer spec files.
        """

        props = {
            'path': self.verifier.paths.root,
            'specs': map(self.verifier.paths.get_path, self.variables.properties[REFERENCE_SPEC_FILES]),
            'filter_fn': self.variables.has_reference,
            'attributes': self.variables.properties[ATTRIBUTES],
            'white_list_patterns': []
        }

        hits = self.seeker.search_specs_for_attributes(props)
        self.assertTrue(len(hits) == 2)

    def test_findAllStringsInSource(self):
        """
        Search source code for izpack string references.
        """
        hits = self.seeker.find_references_in_source(
            patterns=self.strings.properties[PATTERNS],
            path_list=[source_path2],
            white_list_patterns=self.strings.properties[WHITE_LIST_PATTERNS]
        )
        num = len(hits)
        self.assertTrue(num == 8)

    def test_processKeys(self):
        """
        Test the process_keys method.
        """
        # hits list: (key from source, location, expected output)
        hits = [('idata.getString("string.key.1"', 'path/to/file.java', 'string.key.1'),
                ('System.out.println(idata.getString("string.key.2"', 'path/to/file.java', 'string.key.2'),
                ('"string.key.3"', 'path/to/file.java', 'string.key.3'),
                ('somevar + "string.key.4"', 'path/to/file.java', None),
                ('key1', 'data/sample_code_base/src/com/sample/installer/Foo.java', 'some.string.1')
        ]

        seeker = Seeker(None)
        for hit in hits:
            key = seeker.process_key(hit, ['string.in.whitelist'], 'idata.getString\(.*|System.*.println\(.*')
            if key is None:
                print str(hit[2]) + " => " + str(key)
                self.assertEquals(hit[2], key)
            else:
                print str(hit[2]) + " => " + str(key[0])
                self.assertEquals(hit[2], key[0])





__author__ = 'fcanas'


class IzVerifierException(Exception):
    """
    General exception thrown by IzVerifier.
    """

    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


class IzArgumentsException(IzVerifierException):
    """
    Thrown when IzVerifier is called with invalid arguments.
    """


class MissingFileException(IzVerifierException):
    """
    Thrown when a required xml spec or resource file can't be found.
    """




from IzVerifier.izspecs.izproperties import IzProperties

__author__ = 'fcanas'
import unittest

pom = 'data/sample_installer_iz5/pom.xml'
path1 = 'data/sample_installer_iz5/izpack/'
path2 = 'data/sample_installer_iz5/resources/'
source_path2 = 'data/sample_code_base'

class TestSeeker(unittest.TestCase):
    """
    Basic testing of izproperties class.
    """

    def test_parsePom(self):
        """
        Tests izproperties' ability to parse props from a pom file.
        """
        props = IzProperties(pom)
        self.assertEquals(props['izpack.version'], '5.0.0-rc2')
        self.assertEquals(props['project.build.sourceEncoding'], 'UTF-8')


    def test_parseIzProperties(self):
        """
        Test izpack properties file parsing.
        """
        pass

    def test_substituteStrings(self):
        """
        Test the substitution methods.
        """
        props = IzProperties(pom)

        s1 = "${izpack.version}-A.B.C"
        r1 = props.substitute(s1)
        self.assertEquals("5.0.0-rc2-A.B.C", r1)

        s2 = "${project.version}-A.B.C"
        r2 = props.substitute(s2)
        self.assertEquals("5.0.0-rc2-SNAPSHOT-A.B.C", r2)

        s3 = "${undefined.prop}-A.B.C"
        r3 = props.substitute(s3)
        self.assertEquals("${undefined.prop}-A.B.C", r3)

        s4 = "${mistyped.prop-A.B.C"
        r4 = props.substitute(s4)
        self.assertEquals("${mistyped.prop-A.B.C", r4)





from IzVerifier.izspecs.containers.izclasses import IzClasses

__author__ = 'fcanas'

import unittest

from IzVerifier.izspecs.containers.izconditions import IzConditions
from IzVerifier.izspecs.containers.izstrings import IzStrings
from IzVerifier.izspecs.containers.izvariables import IzVariables
from IzVerifier.izverifier import IzVerifier
from IzVerifier.izspecs.containers.constants import *


path1 = 'data/sample_installer_iz5/izpack/'
path2 = 'data/sample_installer_iz5/resources/'
source_path2 = 'data/sample_code_base/src/'
pom = 'data/sample_installer_iz5/pom.xml'


class TestVerifier(unittest.TestCase):
    """
    Basic testing of verifier class.
    """

    def setUp(self):
        args = {
            'specs_path': path1,
            'sources': [source_path2],
            'resources_path': path2,
            'pom': pom,
            'specs': ['conditions', 'strings', 'variables']
        }
        self.izv = IzVerifier(args)
        self.izv.reporter.set_terminal_width()  # Sets width to width of terminal

    def test_IzPaths(self):
        """
        Testing install.xml path parsing.
        """
        specs = [('variables', 'variables.xml'),
                 ('conditions', 'conditions.xml'),
                 ('dynamicvariables', 'dynamic_variables.xml'),
                 ('resources', 'resources.xml'),
                 ('panels', 'panels.xml'),
                 ('packs', 'packs.xml')]

        self.assertTrue(self.izv != None)
        for spec in specs:
            path = self.izv.paths.get_path(spec[0])
            self.assertTrue(spec[1] in path,
                            msg=path + "!=" + spec[1])

    def test_IzConditions(self):
        """
        Testing the strings container.
        """
        conditions = self.izv.paths.get_path('conditions')
        self.assertEquals(conditions, 'data/sample_installer_iz5/izpack/conditions.xml')

        izc = IzConditions(conditions)
        self.assertTrue(izc != None)

        # Test for number of keys in conditions.xml plus white list
        num = len(izc.get_keys()) - len(izc.properties[WHITE_LIST])
        print num
        self.assertEquals(num, 12, str(num) + "!=12")

    def test_langpack_paths(self):
        """
        Test that we parsed the langpack paths from resources.xml
        """
        langpacks = [('default', 'data/sample_installer_iz5/resources/langpacks/CustomLangPack.xml'),
                     ('eng', 'data/sample_installer_iz5/resources/langpacks/CustomLangPack.xml')]

        for tpack, fpack in zip(langpacks, self.izv.paths.get_langpacks().keys()):
            self.assertEquals(tpack[1], self.izv.paths.get_langpack_path(tpack[0]))


    def test_IzStrings(self):
        """
        Testing the strings container.
        """
        langpack = self.izv.paths.get_langpack_path()

        izs = IzStrings(langpack)
        self.assertTrue(izs != None)

        # Test for number of strings
        num = len(izs.get_keys())
        self.assertEquals(num, 4, str(num) + '!=4')

    def test_IzVariables(self):
        """
        Testing the variables container.
        """
        variables = self.izv.paths.get_path('variables')
        self.assertEquals(variables, 'data/sample_installer_iz5/izpack/variables.xml')

        izv = IzVariables(variables)
        self.assertTrue(izv != None)
        num = len(izv.get_keys()) - len(izv.properties[WHITE_LIST])
        self.assertEquals(num, 3, str(num) + '!=3')

    def test_verifyStrings(self):
        """
        Verify strings in sample installer
        """
        hits = self.izv.verify('strings', verbosity=2)
        undefined_strings = {'some.string.4',
                             'my.error.message.id.test',
                             'password.empty',
                             'password.not.equal',
                             'some.user',
                             'some.user.panel.info',
                             'some.user.password',
                             'some.user.password.confirm',
                             'some.string.5',
                             'some.string.6',
                             'hello.world'}

        non_strings = {'db.driver'}

        found_strings, location = zip(*hits)
        for id in undefined_strings:
            self.assertTrue(id in found_strings)

        for id in non_strings:
            self.assertTrue(id not in found_strings)



    def test_verifyConditions(self):
        """
        Verify conditions in sample installer.
        """
        hits = self.izv.verify('conditions', verbosity=2)

        undefined_conditions = {'myinstallerclass.condition',
                                'some.condition.2',
                                'some.condition.1'}

        found_conditions, location = zip(*hits)

        for id in undefined_conditions:
            self.assertTrue(id in found_conditions)

    def test_verifyVariables(self):
        """
        Verify conditions in sample installer.
        """
        hits = self.izv.verify('variables', verbosity=1)
        num = len(hits)
        self.assertTrue(num == 5)

    def test_verifyAll(self):
        """
        Verify all specs on sample installer.
        """
        hits = self.izv.verify_all(verbosity=1)
        num = len(hits)
        assert (num != 0)

    def test_findReference(self):
        """
        Find some references to items in source code and specs.
        """
        hits = self.izv.find_references('some.user.password', verbosity=2)
        self.assertEquals(len(hits), 2)

        hits = self.izv.find_references('password.empty', verbosity=2)
        self.assertEquals(len(hits), 1)

        # Ref in code
        hits = self.izv.find_references('some.string.3', verbosity=2)
        self.assertEquals(len(hits), 1)

        # var substitution not yet implemented for find references, so this
        # test will miss the ref in Foo.java
        hits = self.izv.find_references('some.condition.1', verbosity=2)
        self.assertEquals(len(hits), 1)

    def test_verifyClasses(self):
        """
        Testing the izclasses container.
        """
        classes = IzClasses(source_path2)
        classes.print_keys()
        self.assertEquals(len(classes.get_keys()), 1)

        hits = self.izv.verify('classes', verbosity=2)
        self.assertEquals(len(hits), 5)

        referenced = self.izv.get_referenced('classes')
        self.assertTrue(referenced.has_key('com.sample.installer.Foo'))
        self.assertTrue(referenced.has_key('com.sample.installer.SuperValidator'))
        self.assertTrue(referenced.has_key('com.sample.installer.SuperDuperValidator'))
        self.assertTrue(referenced.has_key('com.sample.installer.BarListener'))


if __name__ == '__main__':
    unittest.main()




from IzVerifier.izspecs.containers.constants import *

__author__ = 'fcanas'


class ConditionDependencyGraph():
    compound_conditions = ['or', 'xor', 'and', 'not']

    def __init__(self, verifier, fail_on_undefined_vars=False):
        self.ill_defined = {}
        self.well_defined = set()
        self.crefs = set((ref[0] for ref in verifier.find_code_references('conditions')))
        self.srefs = set((ref[0] for ref in verifier.find_specification_references('conditions')))
        self.conditions = verifier.get_container('conditions')
        self.variables = verifier.get_container('variables')
        self.drefs = self.conditions.get_keys()
        self.fail_on_undefined_vars = fail_on_undefined_vars

    def all_references(self):
        return self.drefs | self.srefs | self.crefs

    def test_verify_all_dependencies(self):
        """
        For the given installer conditions, verify the dependencies for every single one of the conditions
        that are in some way referenced in specs or source.
        """

        for condition in self.all_references():
            result = self.verify_dependencies(condition)

            if result:
                self.ill_defined[condition] = result
            else:
                self.well_defined.add(condition)

        return self.ill_defined

    def test_verify_dependencies(self, cond_id, conditions, variables):
        """
        Verifies that the given condition id is defined, and that its' dependencies and
        their transitive dependencies are all defined and valid.
        """

        if not cond_id in conditions.get_keys():
            return 1
        else:
            result = self.verify_dependencies(cond_id)
        return result

    def verify_dependencies(self, cond_id):
        """
        Performs a depth-first search of a condition's dependencies in order
        to verify that all dependencies and transitive dependencies are defined
        and valid.
        """
        undefined_paths = set()
        self._verify_dependencies(cond_id, undefined_paths, tuple())
        return undefined_paths

    def _verify_dependencies(self, cond_id, undefined_paths, current_path):
        """
        Given the soup for a condition, test that its dependencies are validly
        defined.
        """
        defined_children = True

        # Exception for izpack conditions:
        if cond_id in self.conditions.properties[WHITE_LIST]:
            return True

        # Short-circuit on well-defined conditions:
        if cond_id in self.well_defined:
            return True

        # Short-circuit ill-defined conditions:
        if cond_id in self.ill_defined.keys():
            current_path = current_path + ((cond_id, 'ill-defined condition'),)
            undefined_paths.add(current_path)
            return False

        # Cycle checking:
        tup = (cond_id, 'condition')
        if tup in current_path:
            current_path += ((cond_id, 'cyclic condition reference'),)
            undefined_paths.add(current_path)
            return False

        # Check for undefined condition.
        if not cond_id in self.conditions.get_keys():
            tup = (cond_id, 'undefined condition')
            current_path += (tup,)
            undefined_paths.add(current_path)
            return False

        current_path += (tup,)
        condition = self.conditions.container[cond_id]
        condition_type = condition['type']

        if 'variable' in condition_type:
            var = str(condition.find('name').text)
            if not var in self.variables.get_keys() and self.fail_on_undefined_vars:
                current_path += ((var, 'undefined variable'),)
                undefined_paths.add(current_path)
                return False

        elif 'exists' in condition_type:
            var = str(condition.find('variable').text)
            if not var in self.variables.get_keys() and self.fail_on_undefined_vars:
                current_path += ((var, 'undefined variable'),)
                undefined_paths.add(current_path)
                return False

        elif condition_type in self.compound_conditions:
            dependencies = condition.find_all('condition')
            for dep in dependencies:
                did = str(dep['refid'])
                if not self._verify_dependencies(did, undefined_paths, current_path):
                    defined_children = False

        if defined_children:
            self.well_defined.add(cond_id)

        return defined_children







from bs4 import BeautifulSoup
from abc import abstractmethod
from IzVerifier.exceptions.IzVerifierException import MissingFileException


class IzContainer():
    """
    Abstract class responsible for containing some izpack spec resource.

    For example implementers, see:
    izconditions (for izpack conditions)
    izstrings (for izpack localized strings)
    izvariables (for izpack variables)
    """

    def __init__(self, path):
        self.container = {}
        self.referenced = {}
        try:
            self.soup = BeautifulSoup(open(path), 'xml')
        except IOError:
            raise MissingFileException("spec not found at: " + path)
        self.parse(self.soup)

    def get_referenced(self):
        """
        Return a dict containing all referenced entities and the location of their references:
        {
            'id1' => [file1, file2, ...],
            ...
        }
        """
        return self.referenced

    @abstractmethod
    def parse(self, soup):
        """
        Parse the xml soup generated from the izpack descriptor file.
        """
        pass

    @abstractmethod
    def get_keys(self):
        """
        Return all unique keys found for the container's entity.
        """
        pass

    @abstractmethod
    def count(self):
        """
        Return a count of all unique keys found for the container's entity.
        """
        pass

    @abstractmethod
    def has_definition(self, element):
        """
        Return true if the given element contains an izpack definition for the container item.
        """
        pass

    @abstractmethod
    def has_reference(self, element):
        """
        Return true if the given element contains an izpack string reference.
        This method is used to define all the rules that allow the verifier to find
        references to the type of izpack entity being searched for.
        """
        pass

    @abstractmethod
    def get_spec_elements(self):
        """
        Returns a set of the elements defining each of the container's entities.
        """
        pass

    @abstractmethod
    def element_sort_key(self):
        """
        Returns the key to use when sorting elements for this container.
        """
        pass



import importlib
from IzVerifier.izspecs.containers.izclasses import IzClasses

from IzVerifier.izspecs.izproperties import IzProperties
from IzVerifier.izspecs.verifiers.dependencies import ConditionDependencyGraph
from IzVerifier.izspecs.verifiers.seeker import Seeker
from IzVerifier.izspecs.containers.constants import *
from IzVerifier.exceptions.IzVerifierException import IzArgumentsException
from IzVerifier.izspecs.izpaths import IzPaths
from IzVerifier.logging.reporter import Reporter


__author__ = 'fcanas'


class IzVerifier():
    """
    Responsible for parsing an IzPack installer's spec files and any user-specified source code,
    and finding inconsistencies with izpack's conditions, variables, strings, process panel jobs,
    etc.
    """

    def __init__(self, args):
        """
        Main entry point for IzVerifier. Args is a dictionary in this form:

        args = {
            'specs_path': path                  # Path to specs folder for installer.
            'resources_path': path              # Path to root resources folder for installer.
            'pom': path                         # Path to pom file, if used for properties.
            'sources': [path1, path2, ...]      # Path(s) to associated source code roots.
        }
        """
        _validate_arguments(args)
        self.reporter = Reporter()
        self.specifications = ['conditions', 'variables', 'strings']
        self.containers = {}
        self.sources = args.get('sources', [])

        if 'pom' in args:
            self.properties = IzProperties(args['pom'])
        else:
            self.properties = None
        self.paths = IzPaths(args['specs_path'], args['resources_path'], self.properties)
        self._fill_classes()
        self.seeker = Seeker(self.paths)

    def verify_all(self, verbosity=0):
        """
        Runs a verification for all specs.
        """
        missing = set([])

        for specification in self.specifications:
            missing |= self.verify(specification, verbosity)
        return missing

    def verify(self, specification, verbosity=0):
        """
        Runs a verification on the given izpack spec: conditions, strings, variables, etc.
        """

        container = self.get_container(specification)
        defined = container.get_keys()
        crefs = self.find_code_references(specification)
        srefs = self.find_specification_references(specification)

        self._load_references(crefs | srefs, container)

        cmissing = _undefined(defined, crefs)
        smissing = _undefined(defined, srefs)

        if verbosity > 0:
            self.reporter.report_test('undefined {0} referenced in code'.format(specification), cmissing)
            self.reporter.report_test('undefined {0} referenced in specs'.format(specification), smissing)

        return cmissing | smissing

    def dependency_verification(self, verbosity=0, fail_on_undefined_vars=False):
        """
        Run a conditions dependency graph search.
        """
        graph = ConditionDependencyGraph(self, fail_on_undefined_vars)

        results = graph.test_verify_all_dependencies()
        if verbosity > 0:
            self.reporter.display_paths(results)
        return results

    def get_container(self, specification):
        """
        Returns an izpack item container filled with specs gathered from the installer
        specified in the constructor.
        """
        if not specification in self.containers:
            return self._init_container(specification)
        else:
            return self.containers[specification]

    def init_container(self, specification):
        """
        Initialize a container to be used for verification.
        """
        module = importlib.import_module("IzVerifier.izspecs.containers.iz" + specification)
        class_ = getattr(module, 'Iz' + specification.title())
        instance = class_(self.paths.get_path(specification))
        self.containers[specification] = instance
        return instance

    def find_code_references(self, specification):
        """
        Find all source code references for specs in each container.
        :param specification: conditions, variables, strings, or other izpack spec
        """
        container = self.get_container(specification)
        hits = self.seeker.find_references_in_source(patterns=container.properties[PATTERNS],
                                                     path_list=self.sources,
                                                     white_list_patterns=container.properties[WHITE_LIST_PATTERNS])

        return hits

    def find_specification_references(self, specification):
        """
        Find all specification xml references for specs in each container.
        :param specification:
        """
        container = self.get_container(specification)

        args = {
            'specs': map(self.paths.get_path, container.properties[REFERENCE_SPEC_FILES]),
            'filter_fn': container.has_reference,
            'attributes': container.properties[ATTRIBUTES],
            'transform_fn': container.ref_transformer,
            'white_list_patterns': container.properties[WHITE_LIST_PATTERNS]
        }
        hits = self.seeker.search_specs_for_attributes(args)
        return hits

    def find_references(self, rid, specs=None, verbosity=0):
        """
        Finds all references to the given id in source code and specs file for any of the given specs.
        """
        results = set()

        if not specs:
            specs = self.specifications

        for spec in specs:
            results |= self.find_reference(spec, rid, verbosity)

        return results

    def find_reference(self, specification, rid, verbosity=0):
        """
        Find all references to the given id in source and specs for the given spec.
        Returns a set of tuple results.
        """
        container = self.get_container(specification)

        props = {
            'path': self.paths.root,
            'id': rid,
            'specs': map(self.paths.get_path, container.properties[REFERENCE_SPEC_FILES]),
            'filter_fn': container.has_reference,
            'attributes': container.properties[ATTRIBUTES],
            'transformer': lambda x: x,
            'patterns': container.properties[PATTERNS],
            'source_paths': self.sources,
            'white_list_patterns': container.properties[WHITE_LIST_PATTERNS]
        }
        results = self.seeker.find_id_references(props)
        if verbosity > 0:
            self.reporter.report_test('references to {0} in {1}'.format(rid, specification), results)
        return results

    @staticmethod
    def _load_references(references, container):
        """
        Load a container's referenced map with all detected references from source code and spec files.
        """
        referenced = container.get_referenced()
        for ref in references:
            if ref[0] in referenced:
                referenced[ref[0]].add(ref[1])
            else:
                referenced[ref[0]] = {ref[1]}

    def _init_container(self, specification):
        """
        Initialize a container to be used for verification.
        """
        module = importlib.import_module("IzVerifier.izspecs.containers.iz" + specification)
        class_ = getattr(module, 'Iz' + specification.title())
        instance = class_(self.paths.get_path(specification))
        self.containers[specification] = instance
        return instance

    def _fill_classes(self):
        """
        Fills a 'classes container' with custom class info from source code paths.
        """
        container = IzClasses()

        for path in self.sources:
            container.parse(path)
        self.containers['classes'] = container

    def get_referenced(self, specification):
        """
        Return a specification's map of id's to references.
        """
        return self.get_container(specification).get_referenced()


def _validate_arguments(args):
    """
    Throws exceptions if required args are missing or invalid.
    """
    if not 'specs_path' in args:
        raise IzArgumentsException("No Path to Installer Specs Specified")
    if not 'resources_path' in args:
        raise IzArgumentsException("No Path to Installer Resources Specified")


def _undefined(key_set, tup_set):
    """
    Returns the subset of keys from tup_set not present in key_set.
    key_set is a simple set of string keys.
    up_set is a set of tuples: tup[0] is the key of that tuple.
    """
    return set((tup for tup in tup_set if not _quote_remover(tup[0]) in key_set))


def _unused(key_set, tup_set):
    """
    Returns the subset of key_set not present in tup_set.
    key_set is a simple set of string keys.
    up_set is a set of tuples: tup[0] is the key of that tuple.
    """
    if not tup_set:
        return key_set
    return set((key for key in key_set if not key in zip(*tup_set)[0]))


def _quote_remover(key):
    """
    Extracts the actual key of the item passed in.
    """
    if key.startswith('"') and key.endswith('"'):
        return key[1:-1]
    if '=' in key:
        key = _quote_remover(''.join(key.split('=')[1:]))
    return key




__author__ = 'fcanas'

from distutils.core import setup

setup(
        name='IzVerifier',
        version='0.0.3.5',
        author='Francisco Canas',
        author_email='mailfrancisco@gmail.com',
        packages=[
            'IzVerifier',
            'IzVerifier.exceptions',
            'IzVerifier.izspecs',
            'IzVerifier.izspecs.containers',
            'IzVerifier.izspecs.verifiers',
            'IzVerifier.test',
            'IzVerifier.logging'
        ],
        url='https://github.com/izv/IzVerifier',
        license='LICENSE.txt',
        description='Static specification file verification for IzPack installers.',
        long_description=open('README.txt').read(),
        install_requires=[
                "beautifulsoup4 >= 4.3.2",
                "lxml >= 3.3.5"
            ]
        )



import subprocess
import re
from bs4 import BeautifulSoup


class Seeker:
    """
    Responsible for finding references to specified izpack entities in an installer's spec xml files
    or in source code.
    """
    grep_location_pattern = "(.*?):"
    grep_comment_pattern = '^\s*[(//)*].+.*$'
    comment_matcher = re.compile(grep_comment_pattern)
    grep_whitelist_patterns = []

    def __init__(self, paths):
        self.paths = paths

    def search_specs_for_attributes(self, args):
        """
        Searches each spec file for any elements that pass the filter_fn, then
        finds occurrences of the given attributes in those elements, Returns a
        set of all values returned by the transform_fn when the found attribute values
        are its parameters.

        args = {
             'path': path to root of installer folder,
             'specs': a list of paths to spec files to search through,
             'filter_fn': a filtering function for xml elements,
             'attributes': a list of the attributes to extract from matching elements,
             'value_fn': a function that filters the extracted values,
             'transform_fn': a function to transform matching values to some form.
        }

         For example:
         in:
         specs = paths to the xml specs we are searching
         filter_fn = element.has_attr('condition') or element.has_attr('conditionid')
         attributes = ['condition','conditionid']
         transform_fn = izconditions.split_compound_conditions

         out:set(all the individual condition ids referenced by spec files, including those
         referenced as compound conditions like condA+condB, etc)

         the default transform_fn is just the identify function: ie it doesn't transform, just
         returns the input value.
        """
        values_found_for_attributes = set()

        specs = args['specs']
        filter_fn = args['filter_fn']
        attributes = args['attributes']
        value_fn = args.get('value_fn', lambda x: True)
        transform_fn = args.get('transform_fn', lambda x: x)
        white_list = args.get('white_list_patterns', [])

        # Search each of the spec files required for elements that pass the filters
        for spec in specs:
            hits = self.search_specs(filter_fn, spec)

            for tag in attributes:
                values = self.extract_attributes(hits, tag)

                # For each found value, transform it and add the result to the set
                for value in values:
                    transformation = transform_fn(value)
                    if isinstance(transformation, set) or isinstance(transformation, list):
                        for transformed_value in transformation:
                            if value_fn(transformed_value):
                                if not self.in_grep_whitelist(transformed_value, white_list):
                                    values_found_for_attributes.add((transformed_value, spec))
                    else:
                        if value_fn(transformation):
                            if not self.in_grep_whitelist(transformation, white_list):
                                values_found_for_attributes.add((transformation, spec))

        return values_found_for_attributes

    def search_specs_for_value(self, args):
        """
        Searches the given spec files for any instances of elements that pass the element_filter
        which hold attributes specified in the specs_list which in turn have the sought after
        value.

        Will return a set of tuples of the form (value, spec file) after transforming the value
        using the transform_fn.

        See description of args above.
        """
        value = args['id']

        def value_matcher(val):
            matcher = re.compile(value)
            return not matcher.match(val) is None

        args['value_fn'] = value_matcher

        hits = self.search_specs_for_attributes(args)
        return hits

    def find_id_references(self, args):
        """
        Performs a search for some entity defined by the search_properties argument.
        Search_properties is a dict, containing the following:

        id              : A string value we are looking for.
        specs           : A list of spec files to search.
        element_matcher : A function that takes an element as input, returns true if
                            this element is a target element.
        attributes      : A list of attributes to search for id in.
        transformer     : A function to transform the value, when found (can be identity function)
        patterns        : A list of grep string patterns for the target in source code, of the form:
                            (find_pattern, key_extract_pattern)
        source paths    : A list of source code paths to search.

        Returns all references found.
        """
        spec_hits = set()

        hits = self.search_specs_for_value(args)
        spec_hits |= hits

        source_hits = self.find_references_in_source(patterns=args['patterns'],
                                                     path_list=args['source_paths'],
                                                     vid=args['id'],
                                                     white_list_patterns=args['white_list_patterns'])

        return spec_hits | source_hits

    def find_references_in_source(self, patterns, path_list, white_list_patterns, vid=None):
        """
        Find all occurrences of these patterns at the source code in the given paths.
        """
        hits = set()
        if len(patterns) != 0:
            searches, extractors = zip(*patterns)
            combined_search_pattern = '|'.join(searches)
            combined_extract_pattern = '|'.join(extractors)

            for path in path_list:
                if vid:
                    search_pattern = combined_search_pattern.format('"' + vid + '"')
                    extract_pattern = combined_extract_pattern.format('"' + vid + '"')
                else:
                    search_pattern = combined_search_pattern.format('.*?')
                    extract_pattern = combined_extract_pattern.format('.*?')

                hits |= self.search_source_for_pattern(path,
                                                       search_pattern,
                                                       extract_pattern,
                                                       white_list_patterns)

        return set(hits)

    @staticmethod
    def match_literal(key):
        # "some literal string";
        literal_string = '^"[\w].*"$'
        literal_matcher = re.compile(literal_string)
        return literal_matcher.match(key)

    @staticmethod
    def match_compound(key):
        # ie. someVariable + "some literal";
        compound_mix = '^[\"\w].*[\+].+[\w].*$'
        compound_matcher = re.compile(compound_mix)
        return compound_matcher.match(key)

    @staticmethod
    def match_variable(key):
        # someVariable;
        variable = '^[\w]+$'
        variable_matcher = re.compile(variable)
        return variable_matcher.match(key)

    @staticmethod
    def match_method(key, pattern):
        method_matcher = re.search(pattern, key)
        return method_matcher

    @staticmethod
    def extract_key_from_method(key):
        extract_method = '\((.*)'
        extract_method_pattern = re.compile(extract_method)
        matched = re.search(extract_method_pattern, key)
        return matched.group(1)

    def process_key(self, key_and_location, white_list, search_pattern):
        """
        Input: a tuple of  (key, file location), and white_list of keys, and a search pattern
        Output: a tuple of a processed key and the location.
        """

        key = key_and_location[0]
        location = key_and_location[1]

        while self.match_method(key, search_pattern):
            key = self.extract_key_from_method(key)

        # this is a tricky key using strings and variables, probably runtime.
        if self.match_compound(key):
            return None
        elif self.match_literal(key):
            key = self.strip_quotes(key)
            return key, location
        # otherwise it's hopefully just a variable, likely runtime, but we can look for it.
        elif self.match_variable(key):
            key = self.find_variable_value(key, location, white_list)
            if key is not None:
                return key, location
        else:
            return None

    @staticmethod
    def strip_quotes(key):
        return key[1:-1]  # strip quotes

    def find_variable_value(self, variable, location, white_list):
        """
        Given a variable and its location, attempt to substitute it for its value. The white_list is
        compared against.
        """
        search_pattern = 'String {0} = \"'.format(variable)
        extract_pattern = 'String {0} = (\".*?\")'.format(variable)

        hits = list(self.search_source_for_pattern(location, search_pattern, extract_pattern, white_list))

        if hits:

            return hits[0][0]  # first hit
        else:
            return None  # unable to id this, so it's runtime.

    def extract_pattern_and_location_from_grep(self, line, extract_pattern):
        """
        Given a line of output from grep, extracts the pattern from the line and
        returns it as  a tuples holding the pattern and its location.
        """

        tuple_ = self.parse_grep_output(line, extract_pattern)
        if not tuple_ is None:
            return tuple_

    def is_valid_output(self, line, white_list):
        """
        Given a line of input from grep, checks to see if line is valid line of code.
        Rejects output in white_list and comments
        """
        if self.is_comment(line):
            return False
        elif self.in_grep_whitelist(line, white_list):
            return False
        else:
            return True

    def is_comment(self, output):
        """
        Filter out any lines that are parts of comments.
        """
        if ':' in output:
            loc, hit = output.split(':', 1)
        else:
            return False

        if self.comment_matcher.match(hit):
            return True
        else:
            return False

    def parse_grep_output(self, output, extract_pattern):
        """
        Given a line of output from a grep output, returns a tuple
        holding the key of the element and the file it was found in.
        """
        match_key = re.search(extract_pattern, output)
        match_loc = re.search(self.grep_location_pattern, output)
        if match_key and match_loc:
            for match in match_key.groups():
                if match is not None:
                    return match, match_loc.group(1)
        if match_key and not match_loc:
            for match in match_key.groups():
                if match is not None:
                    return match, "UNKNOWN"
        else:
            return None

    @staticmethod
    def search_specs(search_fn, path):
        """
        Scrapes the xml spec file given in path and returns a set of all elements meeting the
        search_fn condition.
        """
        try:
            soup = BeautifulSoup(open(path), 'xml')
        except IOError:
            return set()
        return soup.find_all(search_fn)

    @staticmethod
    def extract_attributes(elements, attribute):
        """
        Given a list of html elements, returns a set of the values of the given attribute.
        """
        values = set()
        for ele in elements:
            if ele.has_attr(attribute):
                atty = ele[attribute]
                if type(atty) is list:
                    for a in atty:
                        values.add(a)
                else:
                    values.add(atty)
        return values

    def search_source_for_pattern(self, path, search_pattern, extract_pattern, white_list):
        """
        Searches all files recursively from given path for the search_pattern.
        Returns a set of all lines containing that pattern.
        """
        keys = set()
        cmd_string = "grep -P -R -e '{0}' {1}".format(search_pattern, path)
        cmd = [cmd_string]
        try:
            output = subprocess.check_output(cmd, shell=True)
            for line in output.split("\n"):
                check_line = self.is_valid_output(line, white_list)
                if check_line is False:
                    key_and_location = None
                else:
                    key_and_location = self.extract_pattern_and_location_from_grep(line, extract_pattern)
                if key_and_location is None:
                    continue
                stripped_key_and_location = self.process_key(key_and_location, white_list, search_pattern)
                if stripped_key_and_location is not None:
                    keys.add(stripped_key_and_location)
        except subprocess.CalledProcessError:
            # no hits were found
            pass
        return keys

    def in_grep_whitelist(self, line, white_list_patterns):
        """
        Tests whether an output line from grep matches whitelist patterns,
        for example if it's a commented line.
        """
        for pat in self.grep_whitelist_patterns + white_list_patterns:
            if re.match(pat, line):
                return True
        return False








